(*************************************************************
 *              Auto-assemblage de graphes                   *
 *************************************************************
   Auteur  : Fabien TARISSAN
   Fichier : assembly.ml
   Date    : 09/06/04
 *************************************************************
   Description : Syntaxe abstraite et oprations de bases
                 pour le scnario de construction 
 *************************************************************)

open Agent;;


(* REPRESENTATION DU GRAPHE D'ASSEMBLAGE *)
(*****************************************)

type assembly = {
  init    : graph * role ;  (* graphe initial *)
  final   : graph list;     (* liste de graphes finaux *)

  (* Liste des jonctions binaires autorises *)
  binaire : ((graph * role *  graph * role) * 
	       (graph * (shift * shift))) list;
  
  (* Liste des jonctions internes autorises *)
  interne : ((graph * role * role) 
	     * (graph * shift)) list
}
;;



(* TESTS DE JONCTIONS *)
(**********************)

(* Teste si une jonction interne est autorise *)
let self_join g r1 r2 assem =
  let rec recherche_cycle l =
    match l with
	[] -> None
      | ( (a,b,c),res )::tl ->
	  if (a,b,c) = (g,r1,r2) 
	  then Some(res,List.mem (fst res) assem.final)
	  else if a < g || 
	    (a = g && 
		(b < r1 
		 || (b = r1 && c < r2)))
	  then recherche_cycle tl
	  else None
  in
    recherche_cycle assem.interne
;;



(* Teste si une connexion entre deux composantes *)
(* disjointes est autorise                      *)
let binary_join rev g1 r1 g2 r2 assem = 
  let rec recherche_binaire l =
    match l with
	[] -> None
      | ( (a,b,c,d),(g,(s1,s2)) )::tl ->
	  if (a,b,c,d) = (g1,r1,g2,r2) 
	  then (
	    if rev
	    then Some((g,(s1,s2)),List.mem g assem.final)
	    else Some((g,(s2,s1)),List.mem g assem.final)
	  )
	  else if a < g1 || 
	    (a = g1 && 
		(b < r1  || 
		   (b = r1 && 
		       (c < g2 || 
			  (c = g2 && d < r2)))))
	  then recherche_binaire tl
	  else None
  in
    recherche_binaire assem.binaire
;;




