(*************************************************************
 *              Auto-assemblage de graphes                   *
 *************************************************************
   Auteur  : Fabien TARISSAN
   Fichier : query.ml
   Date    : 09/06/04
 *************************************************************
   Description : Construit l'ensemble de graphes autoriss
 ************************************************************)

open Param;;
open Agent;;
open Assembly;;
open Graphics;;

open Array;;



(* REPRESENTATION DE LA MEMORISATION DES ELEMENTS *)
(**************************************************)

type graph_aut =
    {
      descr  : (int list * (int * int)) array;
      caract : int * int;
      est_final  : bool
    }
;;

let graph_init = {
  descr =  [| [], (size_querywindow_x/2, size_querywindow_y/2) |];
  caract = (1,0);
  est_final = false
}
;;

type graph_concret =
    { adj : int list array;
      taille : int;
      arete : int
    }
;;

type regles_list = graph_aut list
;;

let adj_array el = map fst el.descr
;;

let nb_sommet el = fst el.caract
;;

let nb_arete el = snd el.caract
;;

let to_graph el =
  {
   adj = map fst el.descr;
   taille = nb_sommet el;
   arete = nb_arete el
 }
;;

type resultat = 
    Rem_One of int
  | Rem_All of int
  | Stop
  | New of ((int list * (int * int)) array * (int * int) ) * bool
;;


(* INITIALISATION ET FIN DES FENETRES GRAPHIQUES *)
(*************************************************)


(* ouvre la fentre *)
let init_querygraph () =
  let s = " "^(string_of_int size_querywindow_x)^"x"
    ^(string_of_int size_querywindow_y) in
    open_graph s;
    set_color fond_query;
    fill_rect 0 0 size_querywindow_x size_querywindow_y;
    set_line_width 2;
    set_window_title "Assembly construction"
;;   

(* ferme la fentre *)
let close_querygraph = close_graph
;;



(* AFFICHAGE *)
(*************)

(* dessine un sommet *)
let draw_vert (x,y) = 
  set_color query_vertice;
  fill_circle x y size_querycircle
;;

(* dessine un arc *)
let draw_edg (x1,y1) (x2,y2) =
  set_color query_edges;
  moveto x1 y1;
  lineto x2 y2
;;

(* dessine le graphe *)
let draw_all a = 
  iteri (fun i (l,pos) -> 
	   draw_vert pos;
	   List.iter 
	     ( fun n -> 
		 if n < i then draw_edg pos (snd a.(n));
	     )  l
	) a
;;

(* pour afficher le fait que le graphe est un graph final *)
let mark_as_final () =
  set_color final_mark;
  moveto final_x final_y;
  draw_string "Final"
;;

(* pour supprimer la marque d'un graphe final *)
let clear_final () =
  set_color fond_query;
  moveto final_x final_y;
  draw_string "Final"
;;


(* MANIPULATION DES PERMUTATIONS D'ENTIERS *)
(*******************************************)

let init_permut n = init n (fun x -> x)
;;

let echange t i j =
  let aux = t.(i) in
    t.(i) <- t.(j);
    t.(j) <- aux
;;


let dernier_decroissant t =
  let rec dernier i =
    if i = 0 then 0
    else if t.(i-1) < t.(i) then i
    else dernier (i-1)
  in
    dernier (length t - 1)
;;


let indice_plus_grand t i =
  let v = t.(i-1) in
  let rec indice n =
    if t.(n) > v then n else indice (n+1)
  in
    indice i
;;


let permut_suivante t =
  let i = dernier_decroissant t in
    if i = 0 then false
    else
      let taille = length t in
      let milieu = i + (taille - 1 - i)/2 in
	for x = i to milieu do 
	  echange t x (taille - 1 - x + i)
	done;
	let j = indice_plus_grand t i in
	  echange t (i-1) j;
	  true;
;;


let inv_permut n p =
  let res = ref (-1) in
    iteri (fun i x -> if x = n then res := i) p;
    !res
;;


(* STRUCTURE POUR LES TESTS D'ISOMORPHISME *)
(*******************************************)

type matching_for_isom =
    {
      permut : int array;
      split  : ((int array * int array) * int array) array
    }
;;


let permut_suivante_possible s =
  let rec permut_ieme i =
    if i < 0 then false
    else 
      let p = snd s.split.(i) in
	if permut_suivante (snd s.split.(i))
	then true
	else (
	  iteri (fun i _ -> p.(i) <- i) p;
	  permut_ieme (i-1)
	)
  in
    permut_ieme (length s.split - 1)
;;


let compute_permut s =
  let fct ((a1,a2),p) =
    iteri (fun i num -> s.permut.(num) <- a2.(p.(i))) a1
  in
    iter fct s.split
;;


let prepare_matching liste n = 
  let long = List.length liste in
  let split = make long ( ([||],[||]),[||] ) in

  let structure =
    { permut = make n 0;
      split = split
    }
  in
    
  let rec prepare_struct l i =
    match l with
	[] -> Some(structure)
      | hd::tl ->
	  let (l1,l2) = snd hd in
	  let n = List.length l1 in
	    if n != List.length l2 
	    then None
	    else (
	      split.(i) <- ((of_list l1,of_list l2) , init_permut n);
	      prepare_struct tl (i-1)
	    )
  in
    
    prepare_struct liste (long-1)
;;


let place_taille b long i l =
  let rec place l =
    match l with
	[] -> [long, ( if b then ([i],[]) else ([],[i]) ) ]
      | ((n,(l1,l2)) as hd)::tl ->
	  if n < long then hd::(place tl)
	  else if n = long then (long, if b then (i::l1,l2) 
				 else (l1,i::l2))::tl
	  else (long, if b then ([i],[]) else ([],[i]))::l
  in
    place l
;;


let prepare b a prelist =
  let res = ref prelist in
  let traite_elem b i x =
    let long = List.length x in
      res:= place_taille b long i !res
  in
    iteri (traite_elem b) a;
    !res
;;


(* FONCTIONS DE TESTS D'ISOMORPHISME *)
(*************************************)

let construit_graph g p =
  { adj = mapi (fun x _ -> 
		  List.map (fun i -> p.(i)) 
		    g.adj.(inv_permut x p)) g.adj;
    taille = g.taille;
    arete = g.arete
  }
;;

let egal g1 g2 =
  if g1.taille != g2.taille || g1.arete != g2.arete
  then false
  else 
    let rec verif_elem l1 l2 =
      match l1 with
	  [] -> true
	| hd::tl -> if List.mem hd l2 then verif_elem tl l2 else false
    in
    let rec verif n =
      if n < 0 then true
      else 
	let l1 = g1.adj.(n) and l2 = g2.adj.(n) in
	  if List.length l1 != List.length l2 then false
	  else if verif_elem l1 l2 then verif (n-1) 
	  else false
    in
      verif (g1.taille - 1)
;;


let exist_permut s g reference =
  let rec test () =
    compute_permut s;
    let g2 = construit_graph g s.permut in
      if egal g2 reference
      then true
      else 
	if permut_suivante_possible s
	then test ()
	else false
  in
    test ()
;;



let first_verif g r prep =
  if g.taille != r.taille || g.arete != r.arete
  then None
  else 
    let prep = prepare false r.adj prep in
      prepare_matching prep g.taille
;;


let isomorphe_mem_list g l =
  let prelist = prepare true g.adj [] in
  let rec isomorphe l =
    match l with
	[] -> None
      | (n,hd,_)::tl ->
	  if g.taille > hd.taille then None else
	    match first_verif g hd prelist with
		None -> isomorphe tl
	      | Some(orga) ->
		  if exist_permut orga g hd
		  then Some (n,orga.permut)
		  else isomorphe tl
  in
    isomorphe l
;;

let isomorphe g1 g2 = 
  match isomorphe_mem_list g1 [(1,g2,false)] with
      None -> false
    | Some(_) -> true
;;


let translate_et_ajoute r1 r2 shift i x =
  if i = r1 
  then r2::(List.map (fun n -> n + shift) x)
  else List.map (fun n -> n + shift) x
;;



(* FONCTIONS AUXILIERES *)
(************************)

(* teste si un caractre reprsente un entier *)
let est_entier c =
  let i = int_of_char c - int_of_char '0' in
    (i>=0 && i<10)
;;


(* recherche le sommet correspondant aux coordonnes (x,y) dans l *)
let recherche (x,y) l =
  let num = ref (-1) in
    iteri (fun i (_,(a,b)) ->
	     let dy = (b-y) and dx = (a-x) in
	       if sqrt(float ( dy*dy + dx*dx) ) < float size_querycircle
	       then num := i
	  )
      l;
    !num
;;



let exists_already nouv liste =
  let g = to_graph nouv in
  let (v,e) = nouv.caract in
  let rec exists l =
    match l with
	[] -> false
      | hd::tl ->
	  let (vl,el) = hd.caract in
	    if v < vl || (v = vl && e < el) 
	    then exists tl
	    else 
	      if v = vl && e = el
	      then 
		if isomorphe g (to_graph hd)
		then true
		else exists tl
	      else false
  in
    exists liste
;;

(* place un nouveau graphe dans la liste *)
let rec place nouv liste =
  match liste with
      [] -> [nouv]
    | hd::tl ->
	if nb_sommet nouv < nb_sommet hd 
	  || (nb_sommet nouv = nb_sommet hd && nb_arete nouv 
	      <= nb_arete hd) 
	then 
	  if not (exists_already nouv liste) 
	  then nouv::liste
	  else liste
	else hd::(place nouv tl)
;;


let remove_nth l i =
  let rec rem acc l n =
    if i = n
    then (acc@(List.tl l)) 
    else rem (acc@[List.hd l]) (List.tl l) (n+1)
  in
    rem [] l 0
;;


let remove_all l i =
  let rec rem acc l n =
    if i = n
    then acc 
    else rem (acc@[List.hd l]) (List.tl l) (n+1)
  in
    rem [] l 0
;;

(* CONSTRUCTION DE SOMMETS ET ARETES *)
(*************************************)

(* construit un sommet (et l'affiche) *)
let built_vertice pos l = 
  draw_vert pos;
  append l [| ([],pos) |]
;;


(* construit un arc (et l'affiche) *)
let built_edges pos1 pos2 l =
  let i1 = recherche pos1 l and i2 = recherche pos2 l in
    if (i1 >= 0 && i2 >= 0 && i1 != i2)
    then (
      let (l1,p1) = l.(i1) and (l2,p2) = l.(i2) in
	l.(i1) <- ((i2::l1),p1);
	l.(i2) <- ((i1::l2),p2);
	draw_edg p1 p2;
	true
    )
    else false
;;


(* FONCTIONS PRINCIPALES *)
(*************************)


(* attend une nouvelle instruction (arc ou sommet) *)
let rec wait_new l v e =
  let ev1 = wait_next_event [Button_down;Key_pressed] in
    if ev1.keypressed 
    then 
      if ev1.key = 'e' then (Stop) 
      else  New( (l,(v,e)), ev1.key = 'f' )
    else
      let t1 = Sys.time () in
	while ( button_down () ) do () done;
	let t2 = Sys.time () in
	  if (t2-.t1) < time_pressed
	  then wait_new 
	    (built_vertice (ev1.mouse_x,ev1.mouse_y) l) 
	    (v+1) e
	  else 
	    let b = built_edges 
	      (ev1.mouse_x,ev1.mouse_y) 
	      (mouse_pos ()) l 
	    in
	      wait_new l v (if b then e+1 else e)
;;


(* demande et rajoute un nouveau graphe *)
let get_one ev_init =
  let l = built_vertice (ev_init.mouse_x,ev_init.mouse_y) [| |] in 
    wait_new l 1 0
;;


(* amliore un graphe dj prsent *)
let rec improve l n =
  clear_graph ();
  let rec calc n =
    let ev = wait_next_event [Key_pressed] in
      if not (est_entier ev.key)
      then n
      else 
	let i = int_of_char ev.key - int_of_char '0' in
	  calc (10*n + i)
  in
    
  let n = calc n in
  let t = List.length l in
    
    if n <= 0 || n > t then improve l 0
    else
      let el = List.nth l (n-1) in
      let recup = copy el.descr in
	draw_all recup;
	if el.est_final then mark_as_final ();
	let ev = wait_next_event [Key_pressed] in
	  if est_entier ev.key
	  then 
	    let i = int_of_char ev.key - int_of_char '0' in
	      improve l i
	  else 
	    if ev.key = 'e' then (Rem_One(n-1))
	    else if ev.key = 'a' then (Rem_All(n-1))
	    else if ev.key = 's' then (Stop) 
	    else (
	      clear_final () ;
	      wait_new recup (nb_sommet el) (nb_arete el)
	    )
;;


(* Boucle principale pour obtenir les nouveaux graphes *)
let rec get_new l =
  clear_graph ();
  let ev = wait_next_event [Button_down; Key_pressed] in
    if ev.keypressed && not (est_entier ev.key) then l 
    else
      let res =
	if ev.keypressed
	then improve l (int_of_char ev.key - int_of_char '0') 
	else get_one ev 
      in
	match res with
	    Rem_One(i) ->
	      get_new (remove_nth l i)
	  | Rem_All(i) ->
	      get_new (remove_all l i)
	  | Stop -> get_new l
	  | New(((g,car),b)) -> 
	      let l = place 
		{ descr = g; caract = car; est_final = b} 
		l 
	      in
		get_new l
;;



(* INTERFACES AVEC LE PROGRAMME PRINCIPAL *)
(******************************************)


(* pour construire les rgles d'assemblage *)
let construct_assembly some =
  match some with
      None -> get_new [graph_init]
    | Some(l) -> get_new l
;;




let test_cycle_poss c (n,g) tl t =
  let rec test r1 r2 =
    if r1 < t-1
    then
      if r2 >= t 
      then test (r1+1) (r1+2)
      else
	let adj = mapi (fun i x -> 
			  if i = r1 
			  then r2::x 
			  else 
			    if i = r2 
			    then r1::x 
			    else x) 
	  g.adj 
	in
	let nouv = {adj = adj; taille = t; arete = g.arete + 1} in
	  match isomorphe_mem_list nouv tl with
	      None -> test r1 (r2+1)
	    | Some(res,permut) -> 
		c := !c@[(((n,t),r1,r2),((res,nouv.taille),permut))];
		test r1 (r2+1)
  in
    test 0 1
;;


let test_bin_poss b (n1,g1) liste t1 =
  let rec test (n2,g2) r1 r2 l t2 =
    if r1 <= t1 - 1
    then
      if r2 >= t2
      then test (n2,g2) (r1+1) 0 l t2
      else
	let adj2 = mapi (translate_et_ajoute r2 r1 t1) g2.adj in
 	let adj = append g1.adj adj2 in
	  adj.(r1) <- (r2+g1.taille)::adj.(r1);
 	  let nouv = {adj = adj; taille = t1+t2; 
		      arete = g1.arete+g2.arete+1} 
	  in
	    match isomorphe_mem_list nouv l with
		None -> test (n2,g2) r1 (r2+1) l t2
	      | Some(res,permut) ->
		  let s1 = sub permut 0 t1 in
		  let s2 = sub permut t1 t2 in
		    b := !b@[(((n1,t1),r1,(n2,t2),r2),
			      ((res,nouv.taille),(s1,s2)))];
		    test (n2,g2) r1 (r2+1) l t2
  in
  let rec test_liste l =
    match l with
	[] -> ()
      | (n2,g2,f)::tl ->
	  if f then test_liste tl
	  else ( 
	    test (n2,g2) 0 0 tl g2.taille;
	    test_liste tl
	  )
  in
    test_liste liste
;;



let construct_rules liste =
  let (bin,cy,f) = (ref [], ref [], ref []) in
  let rec construct l =
    match l with
	[] -> (!bin,!cy,!f)
      | (n,g,b)::tl ->
	  if b 
	  then f := !f@[(n,g.taille)]
	  else (
	    test_cycle_poss cy (n,g) tl g.taille;
	    test_bin_poss bin (n,g) l g.taille
	  );
	  construct tl
  in
    construct liste
;;




let to_assembly l = 
  let i = ref 0 in
  let l = List.map 
    (fun x -> i:=!i+1; (!i, to_graph x, x.est_final)) 
    l 
  in
  let (bin,cy,f) = construct_rules l in
    
    {
      init = ((1,1),0);
      final = f;
      binaire = bin;
      interne = cy
    }
;;
