(*************************************************************
 *                Auto-assemblage de graphes                 *
 *************************************************************
   Auteur  : Fabien TARISSAN
   Fichier : main.ml
   Date    : 09/06/04
 *************************************************************
   Description : Programme principal
 ************************************************************)

open Agent
open Assembly;;
open System;;

open Impression;;
open Graphics;;

open Eval;;
open Move;;

open Query;;

open Array;;


let suffixe = "rb";;


(* INITIALISATION *)
(******************)

(* Renvoie un agent  l'tat initial *)
let init_simple_agent x y (g,r) = {
  span = [];
  cycle = [];
  ident = create_com ();
  role = r;
  state = Active(g);

  pos_x = x;
  pos_y = y;

  dx = if Random.int 2 = 0 then Random.float 15. 
  else -. Random.float 10.;
  dy = if Random.int 2 = 0 then Random.float 15. 
  else -. Random.float 10.;
}
;;

(* Pour initialiser l'tat du systme *)
let init_clic param =
  let rec init_rec acc =
    let mouse = wait_next_event [Button_down; Key_pressed] in
      if mouse.keypressed 
      then acc
      else 
	let a = init_simple_agent (float mouse.mouse_x) 
	  (float mouse.mouse_y) param 
	in
	  draw_agent a;
	  init_rec (a::acc)
  in
    init_rec []
;;


(* TRAITEMENT DES PARAMETRES DONNES AU PROGRAMME *)
(*************************************************)

let arg_list = ref [] in
  iteri (fun i x ->
	   if x = "-save" || x = "-s" 
	   then arg_list := ('s',Sys.argv.(i+1))::(!arg_list)
	   else if x = "-pre" || x="-p" 
	   then arg_list := ('p',Sys.argv.(i+1))::(!arg_list)
	   else if x = "-comp" || x="-c" 
	   then arg_list := ('c',Sys.argv.(i+1))::(!arg_list)
	   else if x = "-load" || x="-l" 
	   then arg_list := ('l',Sys.argv.(i+1))::(!arg_list)
	) Sys.argv;


  (* MAIN LOOP *)
  (*************)
  let rec loop_rec sys choice =
    eval sys choice;
    compute_next_move sys;
    let choice = choose_attempt () in
      draw_system sys choice;
      let ev = wait_next_event [Button_down; Key_pressed; Poll] in
	if ev.button
	then 
	  let a = init_simple_agent (float ev.mouse_x)
	    (float ev.mouse_y) sys.assembly.init 
	  in
	    draw_agent a;
	    sys.active <- a::sys.active;
	    loop_rec sys choice
	else
	  if ev.keypressed
	  then 
	    let k = read_key () in
	      if k = 'q'
	      then ()
	      else 
		if k = 'p'
		then (auto_synchronize true; pause sys choice)
		else loop_rec sys choice
	  else loop_rec sys choice

  and pause sys choice =
    let ev = wait_next_event [Button_down; Key_pressed] in
      if ev.keypressed
      then (auto_synchronize false; loop_rec sys choice)
      else
	let a = init_simple_agent (float ev.mouse_x)
	  (float ev.mouse_y) sys.assembly.init 
	in
	  draw_agent a;
	  sys.active <- a::sys.active;
	  pause sys choice
  in


    (* TRAITEMENT DES DONNEES *)
    (**************************)

    if List.mem_assoc 'p' !arg_list && List.mem_assoc 'l' !arg_list
    then failwith "Incompatible options : -load and -pre";


    (* rgles d'assemblage *)
    let a =
      if List.mem_assoc 'p' !arg_list
      then
	let chan = open_in ((List.assoc 'p' !arg_list)^"."^suffixe) 
	in let read = Marshal.from_channel chan in
	  close_in chan;
	  read
      else    
	begin
	  init_querygraph ();
	  let some =
	    if List.mem_assoc 'l' !arg_list 
	    then
	      let chan = open_in (List.assoc 'l' !arg_list) in
	      let read = Marshal.from_channel chan in
		close_in chan;
		Some(read)
	    else None
	  in
	  let res = construct_assembly some in
	    if List.mem_assoc 's' !arg_list
	    then (
	      let chan = open_out_gen
		[Open_trunc;Open_creat;Open_wronly] 0o644 
		(List.assoc 's' !arg_list) 
	      in
		Marshal.to_channel chan res [];
		close_out chan  
	    );
	    close_querygraph ();
	    to_assembly res
	end
    in


      if List.mem_assoc 'c' !arg_list
      then (
	let chan = open_out_gen 
	  [Open_trunc;Open_creat;Open_wronly] 0o644 
	  ((List.assoc 'c' !arg_list)^"."^suffixe)
	in
	  Marshal.to_channel chan a [];
	  close_out chan  
      );


      (* excution du programme *)
      init_graph ();
      let sys = {
	active = init_clic a.init;
	passive = [];
	update = [];
	break = [];
	final = [];
	assembly = a
      } in
	auto_synchronize false;
	loop_rec sys (choose_attempt ());
	finish_graph ()
;;


